/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: xshepherd.c,v $
 *
 * HISTORY
 * $Log: xshepherd.c,v $
 * Revision 1.3  1994/06/11  00:02:25  menze
 * Added copyright notice
 *
 * Revision 1.2  1994/05/15  00:21:35  davidm
 * Support for prolog functions added.
 *
 * Revision 1.1  1994/04/21  05:22:35  davidm
 * Initial revision
 *
 */
/*
 * Input buffer and shepherd thread manager.
 */
#include "xk_assert.h"
#include "xk_debug.h"
#include "romopt.h"
#include "xshepherd.h"
#include "scout_thread.h"

int traceshepherd = 0;

Scout_LIFO	xShepherdThreadCache;
Scout_LIFO	xShepherdBufferCache;
Scout_FIFO	xShepherdFullBuffers;

static int nThreads = XK_SHEPHERD_THREADS;
static int nBufs    = XK_SHEPHERD_BUFFERS;

/*
 * Support for scanning ROM options:
 */
static xkern_return_t	scanThreads (char **, int, int, void*);
static xkern_return_t	scanBufs (char **, int, int, void*);

static RomOpt romOptions[] = {
    { "threads", 3, scanThreads },
    { "buffers", 3, scanBufs }
};


static xkern_return_t
scanThreads(char **str, int nFields, int line, void *arg)
{
    return sscanf(str[2], "%d", &nThreads) == 1 ? XK_SUCCESS : XK_FAILURE;
} /* scanThreads */


static xkern_return_t
scanBufs(char **str, int nFields, int line, void *arg)
{
    return sscanf(str[2], "%d", &nBufs) == 1 ? XK_SUCCESS : XK_FAILURE;
} /* scanBufs */


/*
 * This is where a newly awakened thread starts processing
 * a buffer.  The argument "arg" is a pointer to the buffer
 * to be processed.  Once running, a thread will keep consuming
 * frames from the full buffer queue until it is empty.
 */
static void
processFrame(void *arg)
{
    XShepherdBuffer *buf = (XShepherdBuffer*) arg;

    xTrace2(shepherd, TR_FUNCTIONAL_TRACE,
	    "%s: processing buffer %p",
	    scout_thread_get_name(scout_thread_self), arg);

    xAssert(buf);

    do {
#ifdef XK_SHEPHERD_PROLOG
	if (buf->prolog_fun) {
	    xTrace4(shepherd, TR_EVENTS,
		    "%s: invoking prolog fun %p (buf=%p,arg=%p)",
		    scout_thread_get_name(scout_thread_self),
		    buf->prolog_fun, buf, buf->prolog_arg);

	    (*buf->prolog_fun)(buf, buf->prolog_arg);
	    buf->prolog_fun = 0;
	} /* if */
#endif /* XK_SHEPHERD_PROLOG */

	xDemux(buf->driverProtl, &buf->upmsg);

	/* refresh the message: */
	msgConstructAllocate(&buf->upmsg, XK_MAX_FRAME_SIZE,
			     (char**) &buf->data);

	/* return processed frame to empty buffer list: */
	scout_lifo_push(&xShepherdBufferCache, buf);

	/* get next buffer if there are any: */
	buf = scout_fifo_remove(&xShepherdFullBuffers);
    } while (buf);

    xTrace1(shepherd, TR_FUNCTIONAL_TRACE,
	    "%s: processing completed",
	    scout_thread_get_name(scout_thread_self));

    scout_lifo_push(&xShepherdThreadCache, scout_thread_self());
    /*
     * The continuation function is always "processFrame" and the
     * argument will be set by the invoking driver.
     */
    scout_thread_suspend_with_continuation();
} /* processFrame */


void
xShepherdInit(void)
{
    XShepherdBuffer *bufs;
    Scout_Thread_Options to;
    Scout_Thread *t;
    int i;

    /* read rom: */
    findRomOpts("shepherd",
		romOptions, sizeof(romOptions)/sizeof(romOptions[0]), 0);

    scout_lifo_init(&xShepherdThreadCache);
    scout_lifo_init(&xShepherdBufferCache);
    scout_fifo_init(&xShepherdFullBuffers);

    xTrace2(shepherd, TR_MAJOR_EVENTS,
	    "xShepherdInit: creating %d empty messages of size %d for "
	    "incoming packets", nBufs, XK_MAX_FRAME_SIZE);

    bufs = (XShepherdBuffer*) xMalloc(nBufs * sizeof(XShepherdBuffer));
    for (i = 0; i < nBufs; ++i) {
	bufs[i].prolog_fun = 0;
	msgConstructAllocate(&bufs[i].upmsg, XK_MAX_FRAME_SIZE,
			     (char**)&bufs[i].data);
	scout_lifo_push(&xShepherdBufferCache, &bufs[i]);
    } /* for */

    xTrace1(shepherd, TR_MAJOR_EVENTS,
	    "xShepherdInit: creating %d threads for incoming packets",
	    nThreads);

    to = *scout_thread_default_options;
    to.name = "shepherd";
    for (i = 0; i < nThreads; ++i) {
#ifdef XK_DEBUG
	{
	    char buf[32];

	    sprintf(buf, "shepherd-%d", i + 1);
	    to.name = buf;
	}
#endif /* XK_DEBUG */
	t = scout_thread_create(processFrame, &to);
	scout_lifo_push(&xShepherdThreadCache, t);
    } /* for */
} /* xShepherdInit */

			/*** end of xshepherd.c ***/
