/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_forward.s,v $
 *
 * HISTORY
 * $Log: alpha_forward.s,v $
 * Revision 1.8  1994/06/11  00:02:36  menze
 * Added copyright notice
 *
 * Revision 1.7  1994/04/30  23:32:35  davidm
 * ipl_state array was on entry too small---overwriting "random" variables!
 *
 * Revision 1.6  1994/04/30  00:49:36  davidm
 * Clean up and SCOUT_TICKS_PER_CLOCK replaces hardcode constant 16.
 *
 * Revision 1.5  1994/04/24  03:41:08  davidm
 * relinquish didn't restore t10 and t11 before invoking save_and_replay.
 *
 * Revision 1.4  1994/04/23  04:03:49  davidm
 * Rollforward is now performed at max ipl level.  Otherwise, the
 * interrupted code just gets interrupted immediately again thus
 * entering an endless loop.
 *
 * Revision 1.3  1994/04/23  03:18:18  davidm
 * Added support for SCOUT_PLAY_LEDS.
 *
 * Revision 1.2  1994/04/21  07:12:02  davidm
 * (interrupt_maybe): special cased clock interrupt.  Clock interrupt
 * also resets system LEDs if SCOUT_PLAY_LEDS is defined.
 *
 */
/*
 * Code implementing rollforward.
 */
#include <regdef.h>
#include "alpha_asm.h"
#include "alpha_pal.h"
#include "scout_config.h"
#include "scout_sync_arch.h"
#include "flamingo.h"

	.set noat
 #	.set noreorder

	.extern alpha_interrupt_handler
	.extern interrupt
	.extern scout_age

	.globl interrupt_maybe

	.data

#ifdef TESTING
	.globl current_ipl
	.globl max_pending_ipl
	.globl ipl_state
	.extern pal_rti
	.align 2
current_ipl:
	.long	0

# define SWPIPL \
	lda	t0, current_ipl; \
	ldl	v0, (t0);	 \
	stl	a0, (t0);	 \
	mov	-1, a0;		 \
	mov	-1, t8;		 \
	mov	-1, t9;		 \
	mov	-1, t10;	 \
	mov	-1, t11
# define RTI \
	ldq	a0, PF_PS(sp); \
	lda	a1, current_ipl; \
	stq	a0, (a1); \
	ldq	gp, PF_GP(sp); \
	ldq	fp, PF_PC(sp); /* bad things happen if fp is used! */ \
	ldq	a0, PF_A0(sp); \
	ldq	a1, PF_A1(sp); \
	ldq	a2, PF_A2(sp); \
	lda	sp, PF_SIZE(sp); \
	jmp	(fp)
#else
# define SWPIPL	call_pal PAL_swpipl	# (clobbers v0, a0, t0, t8-t11)
# define RTI	call_pal PAL_rti
#endif

#define IPL_STATE_PENDING	0x00
#define IPL_STATE_A0		0x08
#define IPL_STATE_A1		0x10
#define IPL_STATE_A2		0x18
#define IPL_SIZE		0x20	/* next power of two */
#define IPL_LD_SIZE		5	/* IPL_SIZE = 2^IPL_LD_SIZE */

#define	REG_STATE_V0	0x00
#define	REG_STATE_T0	0x08
#define	REG_STATE_T8	0x10
#define	REG_STATE_T9	0x18
#define	REG_STATE_SIZE	0x20

	.align	3
max_pending_ipl:	.long	-1

	.align 5
reg_state:		.space	REG_STATE_SIZE

	.align 5
ipl_state: 		.space	(SCOUT_MAX_IPL + 1)*IPL_SIZE

	.text

.ent interrupt_maybe 0	# nesting level 0
	.frame sp IF_SIZE ra	# sp is frame pointer, IF_SIZE sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
	#
	# This is the entry point.  Label "relinquish_maybe" should be
	# before this point such that jumps to it will be predicted
	# taken.
	#
	# When getting here, the OSF/1 PALcode has saved off the
	# interrupt frame as described in the Alpha arch. manual
	# already and IPL is set to the interrupt's level (grr...).
	#
interrupt_maybe:
	lda	sp, -IF_SIZE(sp)

	.prologue 1		# end of prologue, yes we use gp

	stq	 v0, IF_V0(sp)	# save regs clobbered by PAL_swpipl
	stq	 t0, IF_T0(sp)	# plus some work space

	/*
	 * Deal with hard clock interrupts quickly:
	 */
	cmpeq	a0, 1, v0	# is it a clock interrupt?
	blbc	v0, full_interrupt	# no ->

	lda	v0, scout_age	# yup, increment scout_age
	ldq	t0, (v0)
	addq	t0, 1, t0
	stq	t0, (v0)
	and	t0, SCOUT_TICKS_PER_CLOCK-1, v0	# (scout_age % 16) == 0?
	beq	v0, play_leds	# yes -> need to do some more work
	ldq	v0, IF_V0(sp)
	ldq	t0, IF_T0(sp)
	lda	sp, IF_SIZE(sp)
	RTI

play_leds:
#ifdef SCOUT_PLAY_LEDS
# if 0
	/* reset system diagnostic LEDs on the back of the system: */
	lda	v0, 0xfffffc0000000000 | SSR
	ldl	t0, (v0)
	andnot	t0, SSR_DIAG_LED_MASK, t0
	stl	t0, (v0)
# else
	/*
	 * Reset bottom right 8 pixels on a PMAG-BA graphics board in
	 * TURBOchannel slot #3 (aka 0, on the DEC 3000/600):
	 */
	and	t0, 0x1f, v0
	bne	v0, full_interrupt
	lda	v0, 0xfffffc01603400f8
	stq	zero, (v0)
# endif
#endif

full_interrupt:
	stq	 t1, IF_T1(sp)
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 a0, IF_T4(sp)	# save a0 in T4 slot
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	t12, IF_T12(sp)

	mov	SCOUT_MAX_IPL, a0	# disable all interrupts
	SWPIPL			# (old ipl in v0, clobbers a0, t0, t8-t11)
end_intr_AS:

	lda	t9, max_pending_ipl
	ldl	t0, (t9)		# if (max_pending_ipl < my_ipl)
	cmplt	t0, v0, t8		#	max_pending_ipl = my_ipl
	cmovlbs	t8, v0, t0
	stl	t0, (t9)

	ldq	a0, PF_PC+IF_SIZE(sp)	# a0 = saved_pc
	lda	t8, interrupt_maybe	# t8 = interrupt_maybe
	lda	t9, end_intr_AS
	cmpule	t8, a0, t10		# t10 = t8 <= saved_pc
	cmpult	a0, t9, t11		# t11 = saved_pc < end_intr_AS
	and	t10, t11, t10		# t10 = saved_pc in intr_AS
	blbs	t10, perform_rollforward

	ldq	t0, IF_T0(sp)		# restore t0 (dest_reg)
	and	t0, 3, t8
	bne	t8, relinquish_maybe	# t0 is unaligned ->

	subq	t0, 4, t0		# t0 = endPC (points to landmark)
	subq	t0, (SCOUT_SYNC_MAX_LENGTH/4), t9	# t9 = endPC - maxlen
	cmpult	a0, t0, t8		# t8 = (saved_pc < endPC)
	cmpule	t9, a0, t9		# t9 = (endPC - MAX_LEN <= saved pc)
	and	t8, t9, t8
	blbs	t8, check_more		# endPC in range ->

relinquish_maybe:
chk_ill:
	lda	t0, max_pending_ipl
	ldl	t0, (t0)		# t0 = max_pending_ipl
	cmplt	v0, t0, t0		# t0 = (my_ipl < max_pending_ipl)
	blbs	t0, relinquish		# yes, need to relinquish ->

	mov	v0, a0			# a0 = my_ipl
	SWPIPL				# (clobbers v0, a0, t0, t8-t11)
	ldq	a0, IF_T4(sp)		# restore a0 (saved in T4's slot)
	stq	t4, IF_T4(sp)
	stq	t5, IF_T5(sp)
	stq	t6, IF_T6(sp)
	stq	t7, IF_T7(sp)

	stq	s0, IF_S0(sp)		# save remaining registers
	stq	s1, IF_S1(sp)
	stq	s2, IF_S2(sp)
	stq	s3, IF_S3(sp)
	stq	s4, IF_S4(sp)
	stq	s5, IF_S5(sp)
	stq	s6, IF_S6(sp)

	stq	a3, IF_A3(sp)
	stq	a4, IF_A4(sp)
	stq	a5, IF_A5(sp)
	stq	ra, IF_RA(sp)
	stq	$at, IF_AT(sp)

	lda	pv, alpha_interrupt_handler
	jsr	ra, (pv), alpha_interrupt_handler

	ldq	 v0, IF_V0(sp)
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t4, IF_T4(sp)
	ldq	 t5, IF_T5(sp)
	ldq	 t6, IF_T6(sp)
	ldq	 t7, IF_T7(sp)
	ldq	 s0, IF_S0(sp)
	ldq	 s1, IF_S1(sp)
	ldq	 s2, IF_S2(sp)
	ldq	 s3, IF_S3(sp)
	ldq	 s4, IF_S4(sp)
	ldq	 s5, IF_S5(sp)
	ldq	 s6, IF_S6(sp)

	ldq	 a3, IF_A3(sp)
	ldq	 a4, IF_A4(sp)
	ldq	 a5, IF_A5(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	t12, IF_T12(sp)
	ldq	 ra, IF_RA(sp)
	ldq	$at, IF_AT(sp)

	lda	sp, IF_SIZE(sp)
	RTI			# return to interrupted process

check_more:
	lda	t11, SCOUT_SYNC_LANDMARK	# t11 = LANDMARK
	ldl	t10, 0(t0)		# t10 = *endPC
	cmpeq	t10, t11, t10		# t10 = (*endPC == LANDMARK)
	blbc	t8, relinquish_maybe	# *endPC != LANDMARK ->

	# check safety of code:

	ldl	t8, (a0)		# insn = *saved_pc
	lda	t3, chk_table
	br	zero, chk_entry		# -> start checking saftey of code

chk_ok:
chk_loop:
	cmpult	a0, t0, t9		# t9 = (pc < endPC)
	blbc	t9, perform_hybrid_rollforward	# code is safe -> rollforward
	.set	noreorder	# the following is hand-optimized
chk_entry:
	zapnot	t8, 0xf, t8		# clear top 32 bits
	addq	a0, 4, a0		# ++pc
	srl	t8, 26, t9		# opCode = next >> 26
	srl	t8, 21, t10		# regA = insn >> 21
	s4addq	t9, t3, t9		# off = 4*opCode + base
	srl	t8, 16, t11		# regB = insn >> 16
	ldl	t9, (t9)		# load jump offset
	sll	t8, 43, t1		# disp = (int)next<<43 (32+11)
	and	t8, 0x1f, t12		# regC = insn & 0x1f
	srl	t8, 12, t2		# shift lit bit (insn<12>) into bit 0
	and	t11, 0x1f, t11		# regB &= 0x1f
	addq	t9, gp, t9
	ldl	t8, 0(a0)		# insn = *pc
	sra	t1, 43, t1		# disp >>= 43 (32+11)
	and	t10, 0x1f, t10		# regA &= 0x1f

	jmp	zero, (t9), chk_iop

chk_iop:
	cmpeq	t10, SCOUT_SYNC_TARGET_REG, t10	# t10 = (regA == target reg)
	cmpeq	t12, SCOUT_SYNC_TARGET_REG, t12	# t12 = (regC == target reg)
	cmpeq	t11, SCOUT_SYNC_TARGET_REG, t11	# t11 = (regB == target reg)
	or	t10, t12, t10	# t10 = (regA == 1) || (regC == 1)
	ornot	t2, t11, t11	# t11 = insn<12> || (regB != 1)
	andnot	t11, t10, t10	# t10 = (insn<12> || (B != 1)) && !(A=1||C=1)
	blbs	t10, chk_loop
	br	zero, relinquish_maybe

chk_mem:
	cmpeq	t10, SCOUT_SYNC_TARGET_REG, t10	# t10 = (regA == target reg)
	cmpeq	t11, SCOUT_SYNC_TARGET_REG, t11 # t11 = (regB == target reg)
	or	t10, t11, t10
	blbc	t10, chk_loop
	br	zero, relinquish_maybe

chk_br:	cmpeq	t10, SCOUT_SYNC_TARGET_REG, t10	# t10 = (regA == target reg)
	blbs	t10, relinquish_maybe
chk_fbr:
	blt	t1, relinquish_maybe	# disp < 0 ->
	s4addq	t1, a0, t1		# t1 = pc + disp
	cmpule	t1, t0, t1		# t1 = pc + disp <= endPC
	blbs	t1, chk_loop		# legal branch ->
	br	zero, relinquish_maybe

	.set	reorder

perform_hybrid_rollforward:
	lda	a0, back_from_hybrid_rollforward
	stq	a0, IF_T0(sp)	# target reg = back_from_hybrid_rollforwar

perform_rollforward:
	#
	# First save own state in ipl_state[my_ipl] so that interrupt
	# can be replayed; GP is the same for all interrupt handlers
	# and need not be saved.
	#
	# v0 = my_ipl
	#
	lda	t1, ipl_state
	sll	v0, IPL_LD_SIZE, t0	# t0 = IPL_SIZE * my_ipl
	addq	t1, t0, t0		# t0 = &ipl_state[my_ipl]
	ldq	a0, IF_T4(sp)		# a0 is backed up by T4 slot
	stq	a0, IPL_STATE_A0(t0)
	stq	a1, IPL_STATE_A1(t0)
	stq	a2, IPL_STATE_A2(t0)
	mov	1, t1			# ipl_state[my_ipl].pending = TRUE
	stq	t1, IPL_STATE_PENDING(t0)

	ldq	t2, (PF_PS+IF_SIZE)(sp)
	lda	t3, reg_state		# save PS in reg_state[0]
	stq	t2, (t3)
	mov	SCOUT_MAX_IPL, t8
	stq	t8, (PF_PS+IF_SIZE)(sp)	# perform rollforward at max ipl
	ldq	 v0, IF_V0(sp)		# restore regs
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	t12, IF_T12(sp)
	lda	sp, IF_SIZE(sp)		# pop stack
	RTI				# perform rollforward
.end interrupt_maybe


.ent back_from_hybrid_rollforward 0	# nesting level 0
	.frame sp IF_SIZE ra	# sp is frame pointer, IF_SIZE sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
	
back_from_hybrid_rollforward:
	lda	sp, -(IF_SIZE+PF_SIZE)(sp)

	stq	 v0, IF_V0(sp)		# save some regs
	stq	 t0, IF_T0(sp)
	# a0 gets restored via PAL_rti
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)

	# build interrupt frame:
	stq	t0, PF_PC+IF_SIZE(sp)	# t0 holds return address (PC)
	stq	gp, PF_GP+IF_SIZE(sp)
	stq	a0, PF_A0+IF_SIZE(sp)
	stq	a1, PF_A1+IF_SIZE(sp)
	stq	a2, PF_A2+IF_SIZE(sp)
	lda	t0, reg_state		# restore PS from reg_state[0]
	ldq	v0, (t0)
	stq	v0, (PF_PS+IF_SIZE)(sp)	# ipl == kernel mode PS
	br	gp, 1f			# load PC into gp
1:	ldgp	gp,(gp)			# now load gp relative to insn's addr
	# FALL THROUGH
	#
	# Copy saved registers from stack to reg_state,
	# scan through ipl_state in priority order and build interrupt
	# stack frames accordingly.  Finally, rti to top stack frame.
	#
	# Registers v0, t0, t8, and t9 are assumed to be
	# saved on the stack.  sp points to the IF structure.
	#
	# This is awkward mostly because PAL_swpipl destroys so many
	# registers.
	#
	# Must be invoked with all interrupts turned off!
	#
	.text
save_and_replay:
	lda	a0, reg_state
	ldq	t8, IF_V0(sp)
	stq	t8, REG_STATE_V0(a0)
	ldq	t0, IF_T0(sp)		# t0 saved in PF_PC slot
	stq	t0, REG_STATE_T0(a0)
	ldq	t8, IF_T8(sp)
	stq	t8, REG_STATE_T8(a0)
	ldq	t9, IF_T9(sp)
	stq	t9, REG_STATE_T9(a0)

	lda	sp, IF_SIZE(sp)		# pop everything except intr. frame
	lda	t8, ipl_state
	lda	t0, max_pending_ipl
	ldl	v0, (t0)		# v0 = max_pending_ipl
mkframes:
	ldq	t9, IPL_STATE_PENDING(t8)
	blbs	t9, push_frame		# pending -> fake PAL frame on stack
mf_loop:
	subq	v0, 1, v0		# --v0
	addq	t8, IPL_SIZE, t8	# ++ipl_state
	bge	v0, mkframes		# if v0 >= 0 ->

	lda	t8, max_pending_ipl
	mov	-1, t9
	stl	t9, (t8)		# reset max_pending_ipl

	ldq	v0, REG_STATE_V0(a0)	# restore remaining registers
	ldq	t0, REG_STATE_T0(a0)
	ldq	t8, REG_STATE_T8(a0)
	ldq	t9, REG_STATE_T9(a0)

	RTI				# and replay interrupts

push_frame:
	lda	sp, -PF_SIZE(sp)	# alloc frame

	lda	t9, max_pending_ipl
	ldl	t9, (t9)
	subq	t9, v0, t9
	stq	t9, PF_PS(sp)		# ipl == kernel mode PS

	lda	t0, interrupt		# replay via normal int handler
	stq	t0, PF_PC(sp)		# in alpha_locore.s
	stq	gp, PF_GP(sp)
	ldq	t9, IPL_STATE_A0(t8)
	stq	t9, PF_A0(sp)
	ldq	t0, IPL_STATE_A1(t8)
	stq	t0, PF_A1(sp)
	ldq	t9, IPL_STATE_A2(t8)
	stq	t9, PF_A2(sp)

	stq	zero, IPL_STATE_PENDING(t8)
	br	zero, mf_loop

	#
	# Relinquish CPU to higher priority interrupt.
	#
	# v0 holds current IPL (my_ipl).
	#
relinquish:
	lda	t1, ipl_state
	sll	v0, IPL_LD_SIZE, t0	# t0 = IPL_SIZE * my_ipl
	addq	t1, t0, t1		# t1 = &ipl_state[my_ipl]

	mov	1, t0
	stq	t0, IPL_STATE_PENDING(t1)
	ldq	t2, IF_T4(sp)		# save interrupt args in
	stq	t2, IPL_STATE_A0(t1)	# ipl_state[my_ipl]
	stq	a1, IPL_STATE_A1(t1)
	stq	a2, IPL_STATE_A2(t1)

	ldq	 t1, IF_T1(sp)		# restore regs that aren't
	ldq	 t2, IF_T2(sp)		# needed anymore
	ldq	 t3, IF_T3(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	t12, IF_T12(sp)
	br	zero, save_and_replay
.end back_from_hybrid_rollforward

	#
	# The following table maps an Alpha opcode into the
	# address of the code that checks whether the instruction
	# is safe to be rolled forward.
	#
	.rdata
chk_table:
	.gprel32 chk_ill	# 0x00 PAL (pal)
	.gprel32 chk_ill	# 0x01 Rsvd
	.gprel32 chk_ill	# 0x02 Rsvd
	.gprel32 chk_ill	# 0x03 Rsvd
	.gprel32 chk_ill	# 0x04 Rsvd
	.gprel32 chk_ill	# 0x05 Rsvd
	.gprel32 chk_ill	# 0x06 Rsvd
	.gprel32 chk_ill	# 0x07 Rsvd
	.gprel32 chk_mem	# 0x08 LDA (mem)
	.gprel32 chk_mem	# 0x09 LDAH (mem)
	.gprel32 chk_ill	# 0x0a Rsvd
	.gprel32 chk_mem	# 0x0b LDQ_U (mem)
	.gprel32 chk_ill	# 0x0c Rsvd
	.gprel32 chk_ill	# 0x0d Rsvd
	.gprel32 chk_ill	# 0x0e Rsvd
	.gprel32 chk_mem	# 0x0f STQ_U (mem)
	.gprel32 chk_iop	# 0x10 INTA (op)
	.gprel32 chk_iop	# 0x11 INTL (op)
	.gprel32 chk_iop	# 0x12 INTS (op)
	.gprel32 chk_iop	# 0x13 INTM (op)
	.gprel32 chk_ill	# 0x14 Rsvd
	.gprel32 chk_ok		# 0x15 FLTV (op)
	.gprel32 chk_ok		# 0x16 FLTI (op)
	.gprel32 chk_ok		# 0x17 FLTL (op)
	.gprel32 chk_mem	# 0x18 MISC (mem)
	.gprel32 chk_ill	# 0x19 PAL (pal)
	.gprel32 chk_ill	# 0x1a JSR (mem)
	.gprel32 chk_ill	# 0x1b PAL (pal)
	.gprel32 chk_ill	# 0x1c Rsvd
	.gprel32 chk_ill	# 0x1d PAL (pal)
	.gprel32 chk_ill	# 0x1e PAL (pal)
	.gprel32 chk_ill	# 0x1f PAL (pal)
	.gprel32 chk_ok		# 0x20 LDF (mem)
	.gprel32 chk_ok		# 0x21 LDG (mem)
	.gprel32 chk_ok		# 0x22 LDS (mem)
	.gprel32 chk_ok		# 0x23 LDT (mem)
	.gprel32 chk_ok		# 0x24 STF (mem)
	.gprel32 chk_ok		# 0x25 STG (mem)
	.gprel32 chk_ok		# 0x26 STS (mem)
	.gprel32 chk_ok		# 0x27 STT (mem)
	.gprel32 chk_mem	# 0x28 LDL (mem)
	.gprel32 chk_mem	# 0x29 LDQ (mem)
	.gprel32 chk_mem	# 0x2a LDL_L (mem)
	.gprel32 chk_mem	# 0x2b LDQ_L (mem)
	.gprel32 chk_mem	# 0x2c STL (mem)
	.gprel32 chk_mem	# 0x2d STQ (mem)
	.gprel32 chk_mem	# 0x2e STL_C (mem)
	.gprel32 chk_mem	# 0x2f STQ_C (mem)
	.gprel32 chk_br		# 0x30 BR (br)
	.gprel32 chk_fbr	# 0x31 FBEQ (br)
	.gprel32 chk_fbr	# 0x32 FBLT (br)
	.gprel32 chk_fbr	# 0x33 FBLE (br)
	.gprel32 chk_br		# 0x34 BSR (br)
	.gprel32 chk_fbr	# 0x35 FBNE (br)
	.gprel32 chk_fbr	# 0x36 FBGE (br)
	.gprel32 chk_fbr	# 0x37 FBGT (br)
	.gprel32 chk_br		# 0x38 BLBC (br)
	.gprel32 chk_br		# 0x39 BEQ (br)
	.gprel32 chk_br		# 0x3a BLT (br)
	.gprel32 chk_br		# 0x3b BLE (br)
	.gprel32 chk_br		# 0x3c BLBS (br)
	.gprel32 chk_br		# 0x3d BNE (br)
	.gprel32 chk_br		# 0x3e BGE (br)
	.gprel32 chk_br		# 0x3f BGT (br)

	# end of table
