/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_prom.c,v $
 *
 * HISTORY
 * $Log: alpha_prom.c,v $
 * Revision 1.3  1994/06/11  00:02:41  menze
 * Added copyright notice
 *
 * Revision 1.2  1994/04/30  23:36:46  davidm
 * Cleanup and RCS header added.
 *
 */
/*
 * Alpha boot PROM interface.
 *
 * This is based on information published in:
 *
 * Alpha Architecture Reference Manual, DEC Order #: EY-L520E-DP.
 *
 * VMS for Alpha Platforms: Internals and Data Structures, Volume 1,
 * Chapter 33 (Bootstrap Processing), DEC Order #: EY-L466E-P1.
 *
 * DEC 3000 300/400/500/600/800 Models, System Programmer's Manual,
 * DEC Part #: EK-D3SYS-PM.A01
 *
 * DECchip 21064-AA Microprocessor Hardware Reference Manual, DEC
 * Order #: EC-N0079-72 (also available online at gatekeeper.dec.com).
 */
#include <machine/rpb.h>
#include "scout_sync_arch.h"
#include "alpha_pal.h"
#include "alpha_misc.h"

static int osf1_pal_code_running = 0;


static long
mfpr_ptbr()
{
    register long v0 asm("0");
    asm volatile("call_pal %0" :: "i"(PAL_mfpr_ptbr) : "0");
    return v0;
} /* mfpr_ptbr */


static long
mfpr_vptb()
{
    register long v0 asm("0");
    asm volatile("call_pal %0" :: "i"(PAL_mfpr_vptb) : "0");
    return v0;
} /* mfpr_vptb */


long prom_dispatcher();		/* reset prototype */

long
prom_dispatcher(long code)
{
    register long v0 asm("0");
    register long pv asm("27");
    long result;
    volatile long saved_code, old_ipl;

    saved_code = code;

    if (osf1_pal_code_running) {
	/* function is not re-entrant */
	old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);
    } /* if */

    pv = crb->rpb_va_disp;		/* va of proc.-value of disp. fun. */
    v0 = ((long*)pv)[1];		/* get entry point from proc.-value */

    asm volatile("mov  %3, $16
		  lda  $9, 0x20000fd8
		  stq $15,  0($9)
		  stq $26,  8($9)
		  stq $29, 16($9)
		  stq $30, 24($9)
		  mov  $9, $30
		  jsr $26, (%0), 0 # %1 %2
		  lda  $9, 0x20000fd8
		  ldq $15,  0($9)
		  ldq $26,  8($9)
		  ldq $29, 16($9)
		  ldq $30, 24($9)"
		 : "r="(v0)
		 : "0"(v0), "r"(pv), "r"(saved_code)
		 : "memory", "9", "10", "11", "12", "13", "14");
    result = v0;
    if (osf1_pal_code_running) {
	scout_swap_ipl(old_ipl);
    } /* if */
    return result;
} /* prom_dispatcher */


/*
 * Switch from VMS to OSF/1 PALcode.  This was reverse-engineered
 * based on the osf_boot loader.
 */
long
prom_switch_to_osf_pal_code(void)
{
    register long sp asm("30"), v0 asm("0");
    register long a0 asm("16"), a1 asm("17"), a2 asm("18");
    register long a3 asm("19"), a4 asm("20");
    long result, saved_regs[2];

    asm volatile ("stq $15, 0(%0)
		   stq $29, 8(%0)" :: "r"(saved_regs) : "memory", "30");

    percpu[rpb->rpb_cpuid].rpb_pcb.rpb_ksp  = sp;
    percpu[rpb->rpb_cpuid].rpb_pcb.rpb_ptbr = mfpr_ptbr();

    a3 = mfpr_vptb();
    a0 = PALvar_OSF1;
    a1 = (long) &&continuation;
    a2 = (long) K0SEG_TO_P((char*)rpb->rpb_selfref + rpb->rpb_percpu_off
			   +  rpb->rpb_cpuid * rpb->rpb_slotsize);
    a4 = sp;
    v0 = -1;
    asm volatile ("call_pal %0 # %1 %2 %3 %4 %5 %6" :
		  : "i"(PAL_mtpr_at),
		    "r"(a0), "r"(a1), "r"(a2), "r"(a3), "r"(a4), "r"(v0)
		  : "memory", "0", "9", "10", "11", "12", "13", "14");
  continuation:
    asm volatile ("ldq $15, 0(%0)
		   ldq $29, 8(%0)" :: "r"(saved_regs) : "15", "26", "29");
    result = v0;
    osf1_pal_code_running = 1;
    return result;
} /* prom_switch_to_osf_pal_code */

			/*** end of prom_dispatcher.c ***/
